from enum import Enum
import pandas as pd
from logging import *


class _ColorType(Enum):
    COLOR_HEX = 1
    COLOR_EN = 2
    COLOR_CN = 3
    COLOR_OCT = 4


def __parse_color_df():
    import os
    folder_path = os.path.dirname(__file__)
    color_csv_file_path = os.path.join(folder_path, "color.csv")
    global __COLOR_DF
    __COLOR_DF = pd.read_csv(color_csv_file_path, index_col=None, header=0)


def __guess_color_type_and_value(color_value: str, color_df: pd.DataFrame) -> (_ColorType, str):
    if "" == color_value.strip():
        return None, None
    # guess Hex
    if color_value.startswith("#"):
        if len(color_value) != 7:
            raise ValueError("IllegalFormatColor" + color_value)
        for char in color_value[1:]:
            if char.lower() not in '0123456789abcdef':
                raise ValueError("IllegalFormatColor " + color_value)
        return _ColorType.COLOR_HEX, str(color_value)

    # try color_en
    color_en_filtered = color_df.loc[color_df['color_en'] == color_value]['color_hex']
    if len(color_en_filtered) > 0:
        return _ColorType.COLOR_EN, color_en_filtered.tolist()[0]

    # try color_cn
    color_cn_filtered = color_df.loc[color_df['color_cn'] == color_value]['color_hex']
    if len(color_cn_filtered) > 0:
        return _ColorType.COLOR_CN, color_cn_filtered.tolist()[0]

    # try oct_values color
    if ',' in color_value:
        try:
            parts = color_value.split(",")
            if len(parts) != 3:
                raise ValueError("IllegalFormatColor " + color_value)

            parsed_int_color = [int(p.strip()) for p in parts]
            if all([0 <= p <= 255 for p in parsed_int_color]):
                return _ColorType.COLOR_OCT, "#" + "".join([hex(p)[2:].upper().zfill(2) for p in parsed_int_color])
            else:
                raise ValueError("IllegalFormatColor " + color_value)

        except Exception as e:
            error(e)
            raise ValueError("IllegalFormatColor " + color_value)

    raise ValueError("Unknown color value: " + color_value)


__COLOR_DF = pd.DataFrame(None, columns=['color_en', 'color_cn', 'color_hex', 'color_oct'])


def translate_color(color_value: str) -> str:
    if len(__COLOR_DF) < 2:
        return color_value
    color_type, translated_color_value = __guess_color_type_and_value(color_value, __COLOR_DF)
    debug("Translate Color Vale " + color_value + " Got Type " + str(color_type) + " Value " + translated_color_value)
    return translated_color_value


init = __parse_color_df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import abc
import inspect
from typing import Union
from types import FunctionType

from aiearth import engine
from aiearth.engine.variable_node import VariableNode
from aiearth.engine.function_node import FunctionNode
from aiearth.engine.customfunction_node import CustomFunctionNode
from aiearth.engine.function_helper import FunctionHelper
from aiearth.core.error.aie_error import AIEError, AIEErrorCode


class Image(FunctionNode):
    def __init__(self, args=None) -> engine.Image:
        if isinstance(args, (int, float, complex)):
            invoke_args = {"value": args}
            super(Image, self).__init__("Image.constant", invoke_args)
        elif isinstance(args, str):
            invoke_args = {"id": args}
            super(Image, self).__init__("Image.load", invoke_args)
        elif isinstance(args, (list, tuple)):
            images = [Image(i) for i in args]
            result = images[0]
            for image in images[1:]:
                invoke_args = {"srcImg": image, "dstImg": result}
                result = FunctionHelper.apply(
                    "Image.addBands", "engine.Image", invoke_args
                )
            super(Image, self).__init__(
                result.func_name, result.invoke_args, result.var_name
            )
        elif args is None:
            image = Image(0)
            invoke_args = {"input": image, "mask": image}
            super(Image, self).__init__("Image.mask", invoke_args)
        elif isinstance(args, engine.variable_node.VariableNode):
            super(Image, self).__init__(args.func_name, args.invoke_args, args.var_name)
        elif isinstance(args, engine.function_node.FunctionNode):
            super(Image, self).__init__(args.func_name, args.invoke_args, args.var_name)
        else:
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"args 只支持number|str|list类型参数, 传入类型为{type(args)}",
            )

    def select(self, bandSelectors: Union[str, list]) -> engine.Image:
        if bandSelectors is not None and not isinstance(bandSelectors, (str, list)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"bandSelectors 只支持(str,list)类型参数, 传入类型为{type(bandSelectors)}",
            )

        if isinstance(bandSelectors, str):
            bandSelectors = [bandSelectors]

        invoke_args = {
            "input": self,
            "bandSelectors": bandSelectors,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "bandSelectors" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数bandSelectors不能为空")

        return FunctionHelper.apply("Image.select", "engine.Image", invoke_args)

    def rename(self, names: list) -> engine.Image:
        if names is not None and not isinstance(names, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"names 只支持list类型参数, 传入类型为{type(names)}"
            )

        invoke_args = {
            "input": self,
            "names": names,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "names" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数names不能为空")

        return FunctionHelper.apply("Image.rename", "engine.Image", invoke_args)

    def normalizedDifference(self, bandNames: list) -> engine.Image:
        if bandNames is not None and not isinstance(bandNames, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"bandNames 只支持list类型参数, 传入类型为{type(bandNames)}",
            )

        invoke_args = {
            "input": self,
            "bandNames": bandNames,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "bandNames" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数bandNames不能为空")

        return FunctionHelper.apply(
            "Image.normalizedDifference", "engine.Image", invoke_args
        )

    def expression(self, expression: str, map: dict = None) -> engine.Image:
        if expression is not None and not isinstance(expression, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"expression 只支持str类型参数, 传入类型为{type(expression)}",
            )

        if map is not None and not isinstance(map, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"map 只支持dict类型参数, 传入类型为{type(map)}"
            )

        invoke_args = {
            "input": self,
            "expression": expression,
            "map": map,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "expression" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数expression不能为空")

        return FunctionHelper.apply("Image.expression", "engine.Image", invoke_args)

    def getMapId(self, vis_params):
        if vis_params is not None and not isinstance(vis_params, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"vis_params 只支持dict类型参数, 传入类型为{type(vis_params)}",
            )
        return engine.client.Maps.getMapId(self, vis_params)

    def getCenter(self) -> tuple:
        bbox = engine.client.InteractiveSession.getBounds(self)
        if bbox is not None and isinstance(bbox, list):
            center = ((bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2)
            return center
        raise AIEError(AIEErrorCode.ARGS_ERROR, f"获取Center失败. bbox: {bbox}")

    def getBounds(self) -> list:
        bbox = engine.client.InteractiveSession.getBounds(self)
        if bbox is not None and isinstance(bbox, list):
            bounds = [bbox[0], bbox[1], bbox[2], bbox[3]]
            return bounds
        raise AIEError(AIEErrorCode.ARGS_ERROR, f"获取Bounds失败. bbox: {bbox}")

    def where(self, test: engine.Image, value: engine.Image) -> engine.Image:
        if test is not None and not isinstance(test, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"test 只支持engine.Image类型参数, 传入类型为{type(test)}"
            )

        if value is not None and not isinstance(value, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"value 只支持engine.Image类型参数, 传入类型为{type(value)}",
            )

        invoke_args = {
            "input": self,
            "test": test,
            "value": value,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "test" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数test不能为空")

        if "value" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数value不能为空")

        return FunctionHelper.apply("Image.where", "engine.Image", invoke_args)

    def abs(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.abs", "engine.Image", invoke_args)

    def acos(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.acos", "engine.Image", invoke_args)

    def add(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.add", "engine.Image", invoke_args)

    def asin(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.asin", "engine.Image", invoke_args)

    def atan(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.atan", "engine.Image", invoke_args)

    def ceil(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.ceil", "engine.Image", invoke_args)

    def clamp(self, low: [int, float], high: [int, float]) -> engine.Image:
        if low is not None and not isinstance(low, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"low 只支持(int,float)类型参数, 传入类型为{type(low)}"
            )

        if high is not None and not isinstance(high, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"high 只支持(int,float)类型参数, 传入类型为{type(high)}"
            )

        invoke_args = {
            "input": self,
            "low": low,
            "high": high,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "low" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数low不能为空")

        if "high" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数high不能为空")

        return FunctionHelper.apply("Image.clamp", "engine.Image", invoke_args)

    def cos(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.cos", "engine.Image", invoke_args)

    def divide(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.divide", "engine.Image", invoke_args)

    def exp(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.exp", "engine.Image", invoke_args)

    def floor(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.floor", "engine.Image", invoke_args)

    def log(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.log", "engine.Image", invoke_args)

    def log10(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.log10", "engine.Image", invoke_args)

    def max(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.max", "engine.Image", invoke_args)

    def min(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.min", "engine.Image", invoke_args)

    def mod(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.mod", "engine.Image", invoke_args)

    def multiply(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.multiply", "engine.Image", invoke_args)

    def pow(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.pow", "engine.Image", invoke_args)

    def round(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.round", "engine.Image", invoke_args)

    def signum(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.signum", "engine.Image", invoke_args)

    def sin(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.sin", "engine.Image", invoke_args)

    def sqrt(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.sqrt", "engine.Image", invoke_args)

    def subtract(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.subtract", "engine.Image", invoke_args)

    def tan(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.tan", "engine.Image", invoke_args)

    def eq(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.eq", "engine.Image", invoke_args)

    def gt(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.gt", "engine.Image", invoke_args)

    def gte(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.gte", "engine.Image", invoke_args)

    def lt(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.lt", "engine.Image", invoke_args)

    def lte(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.lte", "engine.Image", invoke_args)

    def neq(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.neq", "engine.Image", invoke_args)

    def And(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.and", "engine.Image", invoke_args)

    def Not(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.not", "engine.Image", invoke_args)

    def Or(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.or", "engine.Image", invoke_args)

    def bitwiseAnd(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.bitwiseAnd", "engine.Image", invoke_args)

    def bitwiseNot(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.bitwiseNot", "engine.Image", invoke_args)

    def bitwiseOr(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.bitwiseOr", "engine.Image", invoke_args)

    def bitwiseXor(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.bitwiseXor", "engine.Image", invoke_args)

    def leftShift(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.leftShift", "engine.Image", invoke_args)

    def rightShift(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.rightShift", "engine.Image", invoke_args)

    def mask(self, mask: engine.Image = None) -> engine.Image:
        if mask is not None and not isinstance(mask, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"mask 只支持engine.Image类型参数, 传入类型为{type(mask)}"
            )

        invoke_args = {
            "input": self,
            "mask": mask,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.mask", "engine.Image", invoke_args)

    def unmask(
        self, value: engine.Image = None, sameFootprint: bool = True
    ) -> engine.Image:
        if value is not None and not isinstance(value, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"value 只支持engine.Image类型参数, 传入类型为{type(value)}",
            )

        if sameFootprint is not None and not isinstance(sameFootprint, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"sameFootprint 只支持bool类型参数, 传入类型为{type(sameFootprint)}",
            )

        invoke_args = {
            "input": self,
            "value": value,
            "sameFootprint": sameFootprint,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.unmask", "engine.Image", invoke_args)

    def updateMask(self, mask: engine.Image) -> engine.Image:
        if mask is not None and not isinstance(mask, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"mask 只支持engine.Image类型参数, 传入类型为{type(mask)}"
            )

        invoke_args = {
            "image": self,
            "mask": mask,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "mask" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数mask不能为空")

        return FunctionHelper.apply("Image.updateMask", "engine.Image", invoke_args)

    def focalMax(
        self,
        radius: [int, float] = 1.0,
        kernelType: str = "square",
        units: str = "pixels",
        iterations: int = 1,
    ) -> engine.Image:
        if radius is not None and not isinstance(radius, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"radius 只支持(int,float)类型参数, 传入类型为{type(radius)}",
            )

        if kernelType is not None and not isinstance(kernelType, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernelType 只支持str类型参数, 传入类型为{type(kernelType)}",
            )

        if units is not None and not isinstance(units, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"units 只支持str类型参数, 传入类型为{type(units)}"
            )

        if iterations is not None and not isinstance(iterations, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"iterations 只支持int类型参数, 传入类型为{type(iterations)}",
            )

        invoke_args = {
            "input": self,
            "radius": radius,
            "kernelType": kernelType,
            "units": units,
            "iterations": iterations,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.focalMax", "engine.Image", invoke_args)

    def focalMean(
        self,
        radius: [int, float] = 1.0,
        kernelType: str = "square",
        units: str = "pixels",
        iterations: int = 1,
    ) -> engine.Image:
        if radius is not None and not isinstance(radius, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"radius 只支持(int,float)类型参数, 传入类型为{type(radius)}",
            )

        if kernelType is not None and not isinstance(kernelType, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernelType 只支持str类型参数, 传入类型为{type(kernelType)}",
            )

        if units is not None and not isinstance(units, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"units 只支持str类型参数, 传入类型为{type(units)}"
            )

        if iterations is not None and not isinstance(iterations, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"iterations 只支持int类型参数, 传入类型为{type(iterations)}",
            )

        invoke_args = {
            "input": self,
            "radius": radius,
            "kernelType": kernelType,
            "units": units,
            "iterations": iterations,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.focalMean", "engine.Image", invoke_args)

    def focalMedian(
        self,
        radius: [int, float] = 1.0,
        kernelType: str = "square",
        units: str = "pixels",
        iterations: int = 1,
    ) -> engine.Image:
        if radius is not None and not isinstance(radius, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"radius 只支持(int,float)类型参数, 传入类型为{type(radius)}",
            )

        if kernelType is not None and not isinstance(kernelType, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernelType 只支持str类型参数, 传入类型为{type(kernelType)}",
            )

        if units is not None and not isinstance(units, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"units 只支持str类型参数, 传入类型为{type(units)}"
            )

        if iterations is not None and not isinstance(iterations, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"iterations 只支持int类型参数, 传入类型为{type(iterations)}",
            )

        invoke_args = {
            "input": self,
            "radius": radius,
            "kernelType": kernelType,
            "units": units,
            "iterations": iterations,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.focalMedian", "engine.Image", invoke_args)

    def focalMin(
        self,
        radius: [int, float] = 1.0,
        kernelType: str = "square",
        units: str = "pixels",
        iterations: int = 1,
    ) -> engine.Image:
        if radius is not None and not isinstance(radius, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"radius 只支持(int,float)类型参数, 传入类型为{type(radius)}",
            )

        if kernelType is not None and not isinstance(kernelType, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernelType 只支持str类型参数, 传入类型为{type(kernelType)}",
            )

        if units is not None and not isinstance(units, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"units 只支持str类型参数, 传入类型为{type(units)}"
            )

        if iterations is not None and not isinstance(iterations, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"iterations 只支持int类型参数, 传入类型为{type(iterations)}",
            )

        invoke_args = {
            "input": self,
            "radius": radius,
            "kernelType": kernelType,
            "units": units,
            "iterations": iterations,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.focalMin", "engine.Image", invoke_args)

    def focalMode(
        self,
        radius: [int, float] = 1.0,
        kernelType: str = "square",
        units: str = "pixels",
        iterations: int = 1,
    ) -> engine.Image:
        if radius is not None and not isinstance(radius, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"radius 只支持(int,float)类型参数, 传入类型为{type(radius)}",
            )

        if kernelType is not None and not isinstance(kernelType, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernelType 只支持str类型参数, 传入类型为{type(kernelType)}",
            )

        if units is not None and not isinstance(units, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"units 只支持str类型参数, 传入类型为{type(units)}"
            )

        if iterations is not None and not isinstance(iterations, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"iterations 只支持int类型参数, 传入类型为{type(iterations)}",
            )

        invoke_args = {
            "input": self,
            "radius": radius,
            "kernelType": kernelType,
            "units": units,
            "iterations": iterations,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.focalMode", "engine.Image", invoke_args)

    def focalSum(
        self,
        radius: [int, float] = 1.0,
        kernelType: str = "square",
        units: str = "pixels",
        iterations: int = 1,
    ) -> engine.Image:
        if radius is not None and not isinstance(radius, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"radius 只支持(int,float)类型参数, 传入类型为{type(radius)}",
            )

        if kernelType is not None and not isinstance(kernelType, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernelType 只支持str类型参数, 传入类型为{type(kernelType)}",
            )

        if units is not None and not isinstance(units, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"units 只支持str类型参数, 传入类型为{type(units)}"
            )

        if iterations is not None and not isinstance(iterations, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"iterations 只支持int类型参数, 传入类型为{type(iterations)}",
            )

        invoke_args = {
            "input": self,
            "radius": radius,
            "kernelType": kernelType,
            "units": units,
            "iterations": iterations,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.focalSum", "engine.Image", invoke_args)

    def focalStandardDeviation(
        self,
        radius: [int, float] = 1.0,
        kernelType: str = "square",
        units: str = "pixels",
        iterations: int = 1,
    ) -> engine.Image:
        if radius is not None and not isinstance(radius, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"radius 只支持(int,float)类型参数, 传入类型为{type(radius)}",
            )

        if kernelType is not None and not isinstance(kernelType, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernelType 只支持str类型参数, 传入类型为{type(kernelType)}",
            )

        if units is not None and not isinstance(units, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"units 只支持str类型参数, 传入类型为{type(units)}"
            )

        if iterations is not None and not isinstance(iterations, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"iterations 只支持int类型参数, 传入类型为{type(iterations)}",
            )

        invoke_args = {
            "input": self,
            "radius": radius,
            "kernelType": kernelType,
            "units": units,
            "iterations": iterations,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "Image.focalStandardDeviation", "engine.Image", invoke_args
        )

    def focalConway(
        self,
        radius: [int, float] = 1.0,
        kernelType: str = "square",
        units: str = "pixels",
        iterations: int = 1,
    ) -> engine.Image:
        if radius is not None and not isinstance(radius, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"radius 只支持(int,float)类型参数, 传入类型为{type(radius)}",
            )

        if kernelType is not None and not isinstance(kernelType, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernelType 只支持str类型参数, 传入类型为{type(kernelType)}",
            )

        if units is not None and not isinstance(units, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"units 只支持str类型参数, 传入类型为{type(units)}"
            )

        if iterations is not None and not isinstance(iterations, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"iterations 只支持int类型参数, 传入类型为{type(iterations)}",
            )

        invoke_args = {
            "input": self,
            "radius": radius,
            "kernelType": kernelType,
            "units": units,
            "iterations": iterations,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.focalConway", "engine.Image", invoke_args)

    def reduce(self, reducer: engine.Reducer) -> engine.Image:
        if reducer is not None and not isinstance(reducer, engine.Reducer):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"reducer 只支持engine.Reducer类型参数, 传入类型为{type(reducer)}",
            )

        invoke_args = {
            "image": self,
            "reducer": reducer,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "reducer" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数reducer不能为空")

        return FunctionHelper.apply("Image.reduce", "engine.Image", invoke_args)

    def reduceRegion(
        self,
        reducer: engine.Reducer,
        geometry: engine.Geometry = None,
        scale: [int, float] = None,
    ) -> object:
        if reducer is not None and not isinstance(reducer, engine.Reducer):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"reducer 只支持engine.Reducer类型参数, 传入类型为{type(reducer)}",
            )

        if geometry is not None and not isinstance(geometry, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"geometry 只支持engine.Geometry类型参数, 传入类型为{type(geometry)}",
            )

        if scale is not None and not isinstance(scale, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"scale 只支持(int,float)类型参数, 传入类型为{type(scale)}"
            )

        invoke_args = {
            "image": self,
            "reducer": reducer,
            "geometry": geometry,
            "scale": scale,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "reducer" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数reducer不能为空")

        return FunctionHelper.apply("Image.reduceRegion", "object", invoke_args)

    def reduceRegions(
        self,
        collection: engine.FeatureCollection,
        reducer: engine.Reducer,
        scale: [int, float] = None,
    ) -> engine.FeatureCollection:
        if collection is not None and not isinstance(
            collection, engine.FeatureCollection
        ):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"collection 只支持engine.FeatureCollection类型参数, 传入类型为{type(collection)}",
            )

        if reducer is not None and not isinstance(reducer, engine.Reducer):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"reducer 只支持engine.Reducer类型参数, 传入类型为{type(reducer)}",
            )

        if scale is not None and not isinstance(scale, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"scale 只支持(int,float)类型参数, 传入类型为{type(scale)}"
            )

        invoke_args = {
            "image": self,
            "collection": collection,
            "reducer": reducer,
            "scale": scale,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "collection" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数collection不能为空")

        if "reducer" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数reducer不能为空")

        return FunctionHelper.apply(
            "Image.reduceRegions", "engine.FeatureCollection", invoke_args
        )

    def clip(self, geometry: Union[engine.Geometry, engine.Feature]) -> engine.Image:
        if geometry is not None and not isinstance(
            geometry, (engine.Geometry, engine.Feature)
        ):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"geometry 只支持(engine.Geometry,engine.Feature)类型参数, 传入类型为{type(geometry)}",
            )

        invoke_args = {
            "image": self,
            "geometry": geometry,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "geometry" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数geometry不能为空")

        return FunctionHelper.apply("Image.clip", "engine.Image", invoke_args)

    def addBands(
        self, srcImg: engine.Image, names: list = None, overwrite: bool = False
    ) -> engine.Image:
        if srcImg is not None and not isinstance(srcImg, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"srcImg 只支持engine.Image类型参数, 传入类型为{type(srcImg)}",
            )

        if names is not None and not isinstance(names, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"names 只支持list类型参数, 传入类型为{type(names)}"
            )

        if overwrite is not None and not isinstance(overwrite, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"overwrite 只支持bool类型参数, 传入类型为{type(overwrite)}",
            )

        invoke_args = {
            "dstImg": self,
            "srcImg": srcImg,
            "names": names,
            "overwrite": overwrite,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "srcImg" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数srcImg不能为空")

        return FunctionHelper.apply("Image.addBands", "engine.Image", invoke_args)

    @staticmethod
    def constant(value: Union[int, float, list]) -> engine.Image:
        if value is not None and not isinstance(value, (int, float, list)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"value 只支持(int,float,list)类型参数, 传入类型为{type(value)}",
            )

        invoke_args = {
            "value": value,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "value" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数value不能为空")

        return FunctionHelper.apply("Image.constant", "engine.Image", invoke_args)

    def bandNames(self) -> object:
        invoke_args = {
            "image": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.bandNames", "object", invoke_args)

    def bandTypes(self) -> object:
        invoke_args = {
            "image": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.bandTypes", "object", invoke_args)

    def date(self) -> object:
        invoke_args = {
            "image": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.date", "object", invoke_args)

    def get(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "input": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply("Image.get", "object", invoke_args)

    @staticmethod
    def rgb(r: engine.Image, g: engine.Image, b: engine.Image) -> engine.Image:
        if r is not None and not isinstance(r, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"r 只支持engine.Image类型参数, 传入类型为{type(r)}"
            )

        if g is not None and not isinstance(g, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"g 只支持engine.Image类型参数, 传入类型为{type(g)}"
            )

        if b is not None and not isinstance(b, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"b 只支持engine.Image类型参数, 传入类型为{type(b)}"
            )

        invoke_args = {
            "r": r,
            "g": g,
            "b": b,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "r" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数r不能为空")

        if "g" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数g不能为空")

        if "b" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数b不能为空")

        return FunctionHelper.apply("Image.rgb", "engine.Image", invoke_args)

    def rgbToHsv(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.rgbToHsv", "engine.Image", invoke_args)

    def hsvToRgb(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.hsvToRgb", "engine.Image", invoke_args)

    def unitScale(self, low: [int, float], high: [int, float]) -> engine.Image:
        if low is not None and not isinstance(low, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"low 只支持(int,float)类型参数, 传入类型为{type(low)}"
            )

        if high is not None and not isinstance(high, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"high 只支持(int,float)类型参数, 传入类型为{type(high)}"
            )

        invoke_args = {
            "input": self,
            "low": low,
            "high": high,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "low" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数low不能为空")

        if "high" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数high不能为空")

        return FunctionHelper.apply("Image.unitScale", "engine.Image", invoke_args)

    def sinh(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.sinh", "engine.Image", invoke_args)

    def cosh(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.cosh", "engine.Image", invoke_args)

    def tanh(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.tanh", "engine.Image", invoke_args)

    def atan2(self, right: engine.Image) -> engine.Image:
        if right is not None and not isinstance(right, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Image类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply("Image.atan2", "engine.Image", invoke_args)

    def fastAtan2(self, right: engine.Image) -> engine.Image:
        if right is not None and not isinstance(right, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Image类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply("Image.fastAtan2", "engine.Image", invoke_args)

    def convolve(self, kernel: engine.Kernel) -> engine.Image:
        if kernel is not None and not isinstance(kernel, engine.Kernel):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernel 只支持engine.Kernel类型参数, 传入类型为{type(kernel)}",
            )

        invoke_args = {
            "image": self,
            "kernel": kernel,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "kernel" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数kernel不能为空")

        return FunctionHelper.apply("Image.convolve", "engine.Image", invoke_args)

    def toByte(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.toByte", "engine.Image", invoke_args)

    def toInt16(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.toInt16", "engine.Image", invoke_args)

    def toUint16(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.toUint16", "engine.Image", invoke_args)

    def toInt32(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.toInt32", "engine.Image", invoke_args)

    def toUint32(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.toUint32", "engine.Image", invoke_args)

    def toFloat(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.toFloat", "engine.Image", invoke_args)

    def toDouble(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.toDouble", "engine.Image", invoke_args)

    def selfMask(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.selfMask", "engine.Image", invoke_args)

    def polynomial(self, coeffs: list) -> engine.Image:
        if coeffs is not None and not isinstance(coeffs, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"coeffs 只支持list类型参数, 传入类型为{type(coeffs)}"
            )

        invoke_args = {
            "input": self,
            "coeffs": coeffs,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "coeffs" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数coeffs不能为空")

        return FunctionHelper.apply("Image.polynomial", "engine.Image", invoke_args)

    def gamma(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.gamma", "engine.Image", invoke_args)

    def digamma(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.digamma", "engine.Image", invoke_args)

    def remap(
        self,
        fromValue: list,
        toValue: list,
        defaultValue: [int, float] = None,
        bandName: str = None,
    ) -> engine.Image:
        if fromValue is not None and not isinstance(fromValue, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"fromValue 只支持list类型参数, 传入类型为{type(fromValue)}",
            )

        if toValue is not None and not isinstance(toValue, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"toValue 只支持list类型参数, 传入类型为{type(toValue)}"
            )

        if defaultValue is not None and not isinstance(defaultValue, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"defaultValue 只支持(int,float)类型参数, 传入类型为{type(defaultValue)}",
            )

        if bandName is not None and not isinstance(bandName, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"bandName 只支持str类型参数, 传入类型为{type(bandName)}"
            )

        invoke_args = {
            "input": self,
            "fromValue": fromValue,
            "toValue": toValue,
            "defaultValue": defaultValue,
            "bandName": bandName,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "fromValue" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数fromValue不能为空")

        if "toValue" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数toValue不能为空")

        return FunctionHelper.apply("Image.remap", "engine.Image", invoke_args)

    def erf(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.erf", "engine.Image", invoke_args)

    def erfc(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.erfc", "engine.Image", invoke_args)

    def erfInv(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.erfInv", "engine.Image", invoke_args)

    def erfcInv(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.erfcInv", "engine.Image", invoke_args)

    def pixelArea(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.pixelArea", "engine.Image", invoke_args)

    def derivative(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.derivative", "engine.Image", invoke_args)

    def entropy(self, kernel: engine.Kernel) -> engine.Image:
        if kernel is not None and not isinstance(kernel, engine.Kernel):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernel 只支持engine.Kernel类型参数, 传入类型为{type(kernel)}",
            )

        invoke_args = {
            "input": self,
            "kernel": kernel,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "kernel" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数kernel不能为空")

        return FunctionHelper.apply("Image.entropy", "engine.Image", invoke_args)

    def distance(self, kernel: engine.Kernel, skipMasked: bool = True) -> engine.Image:
        if kernel is not None and not isinstance(kernel, engine.Kernel):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernel 只支持engine.Kernel类型参数, 传入类型为{type(kernel)}",
            )

        if skipMasked is not None and not isinstance(skipMasked, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"skipMasked 只支持bool类型参数, 传入类型为{type(skipMasked)}",
            )

        invoke_args = {
            "input": self,
            "kernel": kernel,
            "skipMasked": skipMasked,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "kernel" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数kernel不能为空")

        return FunctionHelper.apply("Image.distance", "engine.Image", invoke_args)

    def spectralDilation(
        self,
        metric: str = "sam",
        kernel: engine.Kernel = None,
        useCentroid: bool = False,
    ) -> engine.Image:
        if metric is not None and not isinstance(metric, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"metric 只支持str类型参数, 传入类型为{type(metric)}"
            )

        if kernel is not None and not isinstance(kernel, engine.Kernel):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernel 只支持engine.Kernel类型参数, 传入类型为{type(kernel)}",
            )

        if useCentroid is not None and not isinstance(useCentroid, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"useCentroid 只支持bool类型参数, 传入类型为{type(useCentroid)}",
            )

        invoke_args = {
            "image": self,
            "metric": metric,
            "kernel": kernel,
            "useCentroid": useCentroid,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "Image.spectralDilation", "engine.Image", invoke_args
        )

    def spectralErosion(
        self,
        metric: str = "sam",
        kernel: engine.Kernel = None,
        useCentroid: bool = False,
    ) -> engine.Image:
        if metric is not None and not isinstance(metric, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"metric 只支持str类型参数, 传入类型为{type(metric)}"
            )

        if kernel is not None and not isinstance(kernel, engine.Kernel):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernel 只支持engine.Kernel类型参数, 传入类型为{type(kernel)}",
            )

        if useCentroid is not None and not isinstance(useCentroid, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"useCentroid 只支持bool类型参数, 传入类型为{type(useCentroid)}",
            )

        invoke_args = {
            "image": self,
            "metric": metric,
            "kernel": kernel,
            "useCentroid": useCentroid,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "Image.spectralErosion", "engine.Image", invoke_args
        )

    def spectralGradient(
        self,
        metric: str = "sam",
        kernel: engine.Kernel = None,
        useCentroid: bool = False,
    ) -> engine.Image:
        if metric is not None and not isinstance(metric, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"metric 只支持str类型参数, 传入类型为{type(metric)}"
            )

        if kernel is not None and not isinstance(kernel, engine.Kernel):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"kernel 只支持engine.Kernel类型参数, 传入类型为{type(kernel)}",
            )

        if useCentroid is not None and not isinstance(useCentroid, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"useCentroid 只支持bool类型参数, 传入类型为{type(useCentroid)}",
            )

        invoke_args = {
            "image": self,
            "metric": metric,
            "kernel": kernel,
            "useCentroid": useCentroid,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "Image.spectralGradient", "engine.Image", invoke_args
        )

    def spectralDistance(
        self, image2: engine.Image, metric: str = "sam"
    ) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        if metric is not None and not isinstance(metric, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"metric 只支持str类型参数, 传入类型为{type(metric)}"
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
            "metric": metric,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply(
            "Image.spectralDistance", "engine.Image", invoke_args
        )

    def lanczos(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.lanczos", "engine.Image", invoke_args)

    def pixelCoordinates(self, projection: str) -> engine.Image:
        if projection is not None and not isinstance(projection, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"projection 只支持str类型参数, 传入类型为{type(projection)}",
            )

        invoke_args = {
            "value": self,
            "projection": projection,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "projection" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数projection不能为空")

        return FunctionHelper.apply(
            "Image.pixelCoordinates", "engine.Image", invoke_args
        )

    def pixelLonLat(self) -> engine.Image:
        invoke_args = {
            "value": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.pixelLonLat", "engine.Image", invoke_args)

    def unmix(self, endmembers: list) -> engine.Image:
        if endmembers is not None and not isinstance(endmembers, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"endmembers 只支持list类型参数, 传入类型为{type(endmembers)}",
            )

        invoke_args = {
            "image": self,
            "endmembers": endmembers,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "endmembers" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数endmembers不能为空")

        return FunctionHelper.apply("Image.unmix", "engine.Image", invoke_args)

    def clipToCollection(self, collection: engine.FeatureCollection) -> engine.Image:
        if collection is not None and not isinstance(
            collection, engine.FeatureCollection
        ):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"collection 只支持engine.FeatureCollection类型参数, 传入类型为{type(collection)}",
            )

        invoke_args = {
            "input": self,
            "collection": collection,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "collection" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数collection不能为空")

        return FunctionHelper.apply(
            "Image.clipToCollection", "engine.Image", invoke_args
        )

    def propertyNames(self) -> object:
        invoke_args = {
            "element": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.propertyNames", "object", invoke_args)

    def copyProperties(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "image1": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.copyProperties", "engine.Image", invoke_args)

    def classify(
        self, classifier: engine.Classifier, outputName: str = "classification"
    ) -> engine.Image:
        if classifier is not None and not isinstance(classifier, engine.Classifier):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"classifier 只支持engine.Classifier类型参数, 传入类型为{type(classifier)}",
            )

        if outputName is not None and not isinstance(outputName, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"outputName 只支持str类型参数, 传入类型为{type(outputName)}",
            )

        invoke_args = {
            "input": self,
            "classifier": classifier,
            "outputName": outputName,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "classifier" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数classifier不能为空")

        return FunctionHelper.apply("Image.classify", "engine.Image", invoke_args)

    def sample(
        self,
        region: engine.Geometry,
        scale: [int, float],
        numPixels: int,
        seed: int = 0,
        geometries: bool = False,
    ) -> engine.FeatureCollection:
        if region is not None and not isinstance(region, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"region 只支持engine.Geometry类型参数, 传入类型为{type(region)}",
            )

        if scale is not None and not isinstance(scale, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"scale 只支持(int,float)类型参数, 传入类型为{type(scale)}"
            )

        if numPixels is not None and not isinstance(numPixels, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"numPixels 只支持int类型参数, 传入类型为{type(numPixels)}"
            )

        if seed is not None and not isinstance(seed, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"seed 只支持int类型参数, 传入类型为{type(seed)}"
            )

        if geometries is not None and not isinstance(geometries, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"geometries 只支持bool类型参数, 传入类型为{type(geometries)}",
            )

        invoke_args = {
            "input": self,
            "region": region,
            "scale": scale,
            "numPixels": numPixels,
            "seed": seed,
            "geometries": geometries,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "region" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数region不能为空")

        if "scale" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数scale不能为空")

        if "numPixels" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数numPixels不能为空")

        return FunctionHelper.apply(
            "Image.sample", "engine.FeatureCollection", invoke_args
        )

    def mask(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.mask", "engine.Image", invoke_args)

    def id(self) -> object:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.id", "object", invoke_args)

    def getNumber(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "input": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply("Image.getNumber", "object", invoke_args)

    def getString(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "input": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply("Image.getString", "object", invoke_args)

    def random(self, seed: int = 0, distribution: str = "uniform") -> engine.Image:
        if seed is not None and not isinstance(seed, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"seed 只支持int类型参数, 传入类型为{type(seed)}"
            )

        if distribution is not None and not isinstance(distribution, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"distribution 只支持str类型参数, 传入类型为{type(distribution)}",
            )

        invoke_args = {
            "input": self,
            "seed": seed,
            "distribution": distribution,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.random", "engine.Image", invoke_args)

    def sampleRegion(
        self, region: engine.Geometry, scale: [int, float], geometries: bool = False
    ) -> engine.FeatureCollection:
        if region is not None and not isinstance(region, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"region 只支持engine.Geometry类型参数, 传入类型为{type(region)}",
            )

        if scale is not None and not isinstance(scale, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"scale 只支持(int,float)类型参数, 传入类型为{type(scale)}"
            )

        if geometries is not None and not isinstance(geometries, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"geometries 只支持bool类型参数, 传入类型为{type(geometries)}",
            )

        invoke_args = {
            "input": self,
            "region": region,
            "scale": scale,
            "geometries": geometries,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "region" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数region不能为空")

        if "scale" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数scale不能为空")

        return FunctionHelper.apply(
            "Image.sampleRegion", "engine.FeatureCollection", invoke_args
        )

    def sampleRegions(
        self,
        collection: engine.FeatureCollection,
        scale: [int, float],
        properties: list = None,
        geometries: bool = False,
    ) -> engine.FeatureCollection:
        if collection is not None and not isinstance(
            collection, engine.FeatureCollection
        ):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"collection 只支持engine.FeatureCollection类型参数, 传入类型为{type(collection)}",
            )

        if scale is not None and not isinstance(scale, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"scale 只支持(int,float)类型参数, 传入类型为{type(scale)}"
            )

        if properties is not None and not isinstance(properties, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"properties 只支持list类型参数, 传入类型为{type(properties)}",
            )

        if geometries is not None and not isinstance(geometries, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"geometries 只支持bool类型参数, 传入类型为{type(geometries)}",
            )

        invoke_args = {
            "input": self,
            "collection": collection,
            "scale": scale,
            "properties": properties,
            "geometries": geometries,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "collection" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数collection不能为空")

        if "scale" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数scale不能为空")

        return FunctionHelper.apply(
            "Image.sampleRegions", "engine.FeatureCollection", invoke_args
        )

    def samplePoints(
        self,
        collection: engine.FeatureCollection,
        scale: [int, float],
        properties: list = None,
        geometries: bool = False,
    ) -> engine.FeatureCollection:
        if collection is not None and not isinstance(
            collection, engine.FeatureCollection
        ):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"collection 只支持engine.FeatureCollection类型参数, 传入类型为{type(collection)}",
            )

        if scale is not None and not isinstance(scale, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"scale 只支持(int,float)类型参数, 传入类型为{type(scale)}"
            )

        if properties is not None and not isinstance(properties, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"properties 只支持list类型参数, 传入类型为{type(properties)}",
            )

        if geometries is not None and not isinstance(geometries, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"geometries 只支持bool类型参数, 传入类型为{type(geometries)}",
            )

        invoke_args = {
            "input": self,
            "collection": collection,
            "scale": scale,
            "properties": properties,
            "geometries": geometries,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "collection" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数collection不能为空")

        if "scale" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数scale不能为空")

        return FunctionHelper.apply(
            "Image.samplePoints", "engine.FeatureCollection", invoke_args
        )

    def cast(self, bandTypes: dict) -> engine.Image:
        if bandTypes is not None and not isinstance(bandTypes, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"bandTypes 只支持dict类型参数, 传入类型为{type(bandTypes)}",
            )

        invoke_args = {
            "input": self,
            "bandTypes": bandTypes,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "bandTypes" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数bandTypes不能为空")

        return FunctionHelper.apply("Image.cast", "engine.Image", invoke_args)

    def set(self, var_args: dict) -> engine.Image:
        if var_args is not None and not isinstance(var_args, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"var_args 只支持dict类型参数, 传入类型为{type(var_args)}"
            )

        invoke_args = {
            "input": self,
            "var_args": var_args,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "var_args" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数var_args不能为空")

        return FunctionHelper.apply("Image.set", "engine.Image", invoke_args)

    def glcmTexture(self, size: int = 1) -> engine.Image:
        if size is not None and not isinstance(size, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"size 只支持int类型参数, 传入类型为{type(size)}"
            )

        invoke_args = {
            "input": self,
            "size": size,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.glcmTexture", "engine.Image", invoke_args)

    def zeroCrossing(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.zeroCrossing", "engine.Image", invoke_args)

    def pixelMatrixMultiply(
        self,
        image2: engine.Image,
        leftRows: int,
        leftCols: int,
        rightRows: int,
        rightCols: int,
    ) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        if leftRows is not None and not isinstance(leftRows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftRows 只支持int类型参数, 传入类型为{type(leftRows)}"
            )

        if leftCols is not None and not isinstance(leftCols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftCols 只支持int类型参数, 传入类型为{type(leftCols)}"
            )

        if rightRows is not None and not isinstance(rightRows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rightRows 只支持int类型参数, 传入类型为{type(rightRows)}"
            )

        if rightCols is not None and not isinstance(rightCols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rightCols 只支持int类型参数, 传入类型为{type(rightCols)}"
            )

        invoke_args = {
            "input": self,
            "image2": image2,
            "leftRows": leftRows,
            "leftCols": leftCols,
            "rightRows": rightRows,
            "rightCols": rightCols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        if "leftRows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数leftRows不能为空")

        if "leftCols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数leftCols不能为空")

        if "rightRows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rightRows不能为空")

        if "rightCols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rightCols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixMultiply", "engine.Image", invoke_args
        )

    def pixelMatrixTranspose(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixTranspose", "engine.Image", invoke_args
        )

    def pixelMatrixPseudoInverse(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixPseudoInverse", "engine.Image", invoke_args
        )

    def pixelMatrixInverse(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixInverse", "engine.Image", invoke_args
        )

    def pixelMatrixCholesky(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixCholesky", "engine.Image", invoke_args
        )

    def pixelMatrixSingularValueDecomposition(
        self, rows: int, cols: int
    ) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixSingularValueDecomposition", "engine.Image", invoke_args
        )

    def pixelMatrixQRDecomposition(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixQRDecomposition", "engine.Image", invoke_args
        )

    def pixelMatrixLUDecomposition(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixLUDecomposition", "engine.Image", invoke_args
        )

    def pixelMatrixTrace(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixTrace", "engine.Image", invoke_args
        )

    def pixelMatrixGramian(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixGramian", "engine.Image", invoke_args
        )

    def pixelMatrixDeterminant(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixDeterminant", "engine.Image", invoke_args
        )

    def pixelMatrixSolve(
        self,
        image2: engine.Image,
        leftRows: int,
        leftCols: int,
        rightRows: int,
        rightCols: int,
    ) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        if leftRows is not None and not isinstance(leftRows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftRows 只支持int类型参数, 传入类型为{type(leftRows)}"
            )

        if leftCols is not None and not isinstance(leftCols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftCols 只支持int类型参数, 传入类型为{type(leftCols)}"
            )

        if rightRows is not None and not isinstance(rightRows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rightRows 只支持int类型参数, 传入类型为{type(rightRows)}"
            )

        if rightCols is not None and not isinstance(rightCols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rightCols 只支持int类型参数, 传入类型为{type(rightCols)}"
            )

        invoke_args = {
            "input": self,
            "image2": image2,
            "leftRows": leftRows,
            "leftCols": leftCols,
            "rightRows": rightRows,
            "rightCols": rightCols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        if "leftRows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数leftRows不能为空")

        if "leftCols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数leftCols不能为空")

        if "rightRows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rightRows不能为空")

        if "rightCols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rightCols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixSolve", "engine.Image", invoke_args
        )

    def pixelMatrixIdentity(self, size: int) -> engine.Image:
        if size is not None and not isinstance(size, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"size 只支持int类型参数, 传入类型为{type(size)}"
            )

        invoke_args = {
            "input": self,
            "size": size,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "size" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数size不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixIdentity", "engine.Image", invoke_args
        )

    def pixelMatrixFnorm(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixFnorm", "engine.Image", invoke_args
        )

    def pixelMatrixEigen(self, rows: int, cols: int) -> engine.Image:
        if rows is not None and not isinstance(rows, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"rows 只支持int类型参数, 传入类型为{type(rows)}"
            )

        if cols is not None and not isinstance(cols, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"cols 只支持int类型参数, 传入类型为{type(cols)}"
            )

        invoke_args = {
            "input": self,
            "rows": rows,
            "cols": cols,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "rows" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数rows不能为空")

        if "cols" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数cols不能为空")

        return FunctionHelper.apply(
            "Image.pixelMatrixEigen", "engine.Image", invoke_args
        )

    def hypot(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "input": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.hypot", "engine.Image", invoke_args)

    def interpolate(
        self, axisX: list, axisY: list, behavior: str = "extrapolate"
    ) -> engine.Image:
        if axisX is not None and not isinstance(axisX, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"axisX 只支持list类型参数, 传入类型为{type(axisX)}"
            )

        if axisY is not None and not isinstance(axisY, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"axisY 只支持list类型参数, 传入类型为{type(axisY)}"
            )

        if behavior is not None and not isinstance(behavior, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"behavior 只支持str类型参数, 传入类型为{type(behavior)}"
            )

        invoke_args = {
            "input": self,
            "axisX": axisX,
            "axisY": axisY,
            "behavior": behavior,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "axisX" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数axisX不能为空")

        if "axisY" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数axisY不能为空")

        return FunctionHelper.apply("Image.interpolate", "engine.Image", invoke_args)

    def first(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "input": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.first", "engine.Image", invoke_args)

    def firstNonZero(self, image2: engine.Image) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        invoke_args = {
            "input": self,
            "image2": image2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.firstNonZero", "engine.Image", invoke_args)

    def bitCount(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.bitCount", "engine.Image", invoke_args)

    def blend(self, image2: engine.Image, alpha: [int, float] = 1.0) -> engine.Image:
        if image2 is not None and not isinstance(image2, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"image2 只支持engine.Image类型参数, 传入类型为{type(image2)}",
            )

        if alpha is not None and not isinstance(alpha, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"alpha 只支持(int,float)类型参数, 传入类型为{type(alpha)}"
            )

        invoke_args = {
            "input": self,
            "image2": image2,
            "alpha": alpha,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "image2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数image2不能为空")

        return FunctionHelper.apply("Image.blend", "engine.Image", invoke_args)

    def cbrt(self) -> engine.Image:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Image.cbrt", "engine.Image", invoke_args)

    def rasterLocalUDF(self, udf: object, others: list, params: dict) -> engine.Image:
        if udf is not None and not isinstance(udf, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"udf 只支持object类型参数, 传入类型为{type(udf)}"
            )

        if others is not None and not isinstance(others, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"others 只支持list类型参数, 传入类型为{type(others)}"
            )

        if params is not None and not isinstance(params, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"params 只支持dict类型参数, 传入类型为{type(params)}"
            )

        invoke_args = {
            "input": self,
            "udf": udf,
            "others": others,
            "params": params,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "udf" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数udf不能为空")

        if "others" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数others不能为空")

        if "params" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数params不能为空")

        return FunctionHelper.apply("Image.rasterLocalUDF", "engine.Image", invoke_args)

    def rasterZonalUDAF(
        self,
        udaf: object,
        geometry: engine.Geometry = None,
        scale: [int, float] = 1000.0,
        params: dict = None,
    ) -> object:
        if udaf is not None and not isinstance(udaf, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"udaf 只支持object类型参数, 传入类型为{type(udaf)}"
            )

        if geometry is not None and not isinstance(geometry, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"geometry 只支持engine.Geometry类型参数, 传入类型为{type(geometry)}",
            )

        if scale is not None and not isinstance(scale, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"scale 只支持(int,float)类型参数, 传入类型为{type(scale)}"
            )

        if params is not None and not isinstance(params, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"params 只支持dict类型参数, 传入类型为{type(params)}"
            )

        invoke_args = {
            "input": self,
            "udaf": udaf,
            "geometry": geometry,
            "scale": scale,
            "params": params,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "udaf" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数udaf不能为空")

        return FunctionHelper.apply("Image.rasterZonalUDAF", "object", invoke_args)



#!/usr/bin/env python

import re
from setuptools import setup, find_packages, find_namespace_packages

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


def GetVersion():
    with open("aiearth/engine/__init__.py") as f:
        return re.findall(r"__version__\s*=\s*\"([.\d]+)\"", f.read())[0]


__version__ = GetVersion()
requirements = open("requirements.txt").readlines()

# packages = find_packages(exclude=["tests"])
packages=find_namespace_packages(include=['aiearth.*'], exclude=['test', 'tests'])
setup(
    name="aiearth-engine",
    version=__version__,
    description="AIEarth Engine Python SDK Engine",
    url="https://engine-aiearth.aliyun.com/",
    packages=packages,
    python_requires=">=3.8.0",
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type='text/markdown',
    include_package_data=True
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from aiearth.core.error import AIEError,AIEErrorCode

class Optimizer(object):
    def __init__(self, result, values, use_count):
        # ref
        self.result = result
        # [(ref, value)]
        self.values = values
        # {ref: use_count}
        self.use_count = use_count

        # {ref: optimized_ref}
        self.optimized_ref_map = {}
        # {ref: value}
        self.optimized_values = {}

    def optimize(self):
        optimized_result = self.optimize_referred_value(self.result)
        return optimized_result

    def optimize_value(self, value):
        if any(x in value for x in ["constantValue", "argumentReference"]):
            return value
        elif "arrayValue" in value:
            node = value["arrayValue"]
            optimized_value = [self.optimize_value(v) for v in node["values"]]

            if all("constantValue" in value for value in optimized_value):
                return {
                    "constantValue": [value["constantValue"] for value in optimized_value]
                }
            else:
                return {"arrayValue": {"values": optimized_value}}
        elif "dictionaryValue" in value:
            node = value["dictionaryValue"]
            optimized_value = {k: self.optimize_value(
                v) for k, v in node["values"].items()}
            if all("constantValue" in value for value in optimized_value.values()):
                return {
                    "constantValue": {key: value["constantValue"] for key, value in optimized_value.items()}
                }
            else:
                return {"dictionaryValue": {"values": optimized_value}}
        elif "functionInvocationValue" in value:
            node = value["functionInvocationValue"]

            optimized_value = {}
            optimized_value["functionName"] = node["functionName"]
            optimized_value["arguments"] = {k: self.optimize_value(
                v)for k, v in node["arguments"].items()}
            return {"functionInvocationValue": optimized_value}
        elif "valueReference" in value:
            ref = value["valueReference"]
            if self.use_count[ref] == 1:
                referenced_value = self.values[int(ref)][1]
                return self.optimize_value(referenced_value)
            else:
                return {"valueReference": self.optimize_referred_value(ref)}
        elif "functionDefinitionValue" in value:
            node = value["functionDefinitionValue"]

            optimized_value = {}
            optimized_value["argumentNames"] = node["argumentNames"]
            optimized_value["body"] = self.optimize_referred_value(
                node["body"])
            return {
                "functionDefinitionValue": optimized_value
            }
        else:
            raise AIEError(AIEErrorCode.DEFAULT_INTERNAL_ERROR, "",
                           f"optimize不支持value: {value}")

    def optimize_referred_value(self, reference):
        if reference in self.optimized_ref_map:
            return self.optimized_ref_map[reference]
        optimized_ref = str(len(self.optimized_ref_map))
        self.optimized_ref_map[reference] = optimized_ref
        self.optimized_values[optimized_ref] = self.optimize_value(
            self.values[int(reference)][1])
        return optimized_ref



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import abstractmethod, ABCMeta
from aiearth.engine.client.interactive_session import InteractiveSession


class FunctionNode(metaclass=ABCMeta):
    def __init__(self, name, args, var_name=None):
        self.func_name = name
        self.invoke_args = args
        self.var_name = var_name

    def _isVariable(self):
        return self.func_name is None and self.invoke_args is None

    def getInfo(self):
        return InteractiveSession.getInfo(self)



import hashlib
# import inspect
import json
import pickle
from aiearth.core import g_var

import logging

from aiearth.core.client.client import BaseClient
from aiearth.core.client.endpoints import Endpoints
from aiearth.core.error import AIEError
from oss2 import StsAuth, Bucket
from tqdm import tqdm

from aiearth.engine.serialize import inspect
from aiearth.engine.udf_base.annotation import annotation_parser

log_lvl = 'INFO'
if g_var.has_var(g_var.GVarKey.Log.LOG_LEVEL):
    log_lvl = g_var.get_var(g_var.GVarKey.Log.LOG_LEVEL).upper()
logging.basicConfig(level=log_lvl)
logger = logging.getLogger(__name__)

IMPORTS = """
from aiearth.engine.udf_base.aie_udf_base import *
from aiearth.engine.udf_base.aie_udaf_base import AIEUDAF
from aiearth.engine.udf_base.annotation import *
from aiearth.engine.udf_base.images import *
"""


class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""

    def __init__(self, desc: str = None):
        super().__init__(unit='B', unit_divisor=1024, miniters=1, unit_scale=True, desc=desc)

    def update_to(self, byte_consumed, total_bytes):
        if total_bytes is not None:
            self.total = total_bytes
        return self.update(byte_consumed - self.n)  # also sets self.n = b * bsize


def convert_to_bytes(obj):
    # 将对象转换为字节流
    byte_stream = pickle.dumps(obj)

    return byte_stream


def calculate_md5(obj_bytes):
    # 计算MD5哈希值
    md5_hash = hashlib.md5()
    md5_hash.update(obj_bytes)
    md5_value = md5_hash.hexdigest()

    return md5_value


def get_source(udf_class):
    func_source = inspect.getsource(udf_class)
    return IMPORTS + func_source


def get_meta(source):
    return annotation_parser(source)


def manage(udf_class):
    # get md5 from func
    func_source: str = get_source(udf_class)
    func_uuid = calculate_md5(bytes(func_source, encoding='utf-8'))
    udf_name = udf_class.__name__

    resp = BaseClient.get(f"{Endpoints.HOST}/mariana/openapi/oss/getUdfStsToken?uuid={func_uuid}", {})
    if resp.status_code != 200:
        raise AIEError(10000000, f"注册{udf_name}时遇到了错误: {resp.content.decode(encoding='utf-8')}")
    result = json.loads(resp.content.decode(encoding='utf-8'))
    success = result['success']
    if not success:
        if result['code'] == 24240405:
            # udf 已经上传完成
            return func_uuid
        raise AIEError(10000000, f'注册{udf_name}时遇到了错误: {result["message"]}')
    logger.info(f'Registering UDF {udf_name}')
    module = result['module']
    if not isinstance(module, list) or len(module) < 2:
        raise AIEError(10000000, f"注册{udf_name}时遇到了错误: {resp.content.decode(encoding='utf-8')}")
    pickle_file_key_sts = module[0]
    meta_file_key_sts = module[1]

    auth1 = StsAuth(access_key_id=pickle_file_key_sts['accessKeyId'],
                    access_key_secret=pickle_file_key_sts['accessKeySecret'],
                    security_token=pickle_file_key_sts['securityToken'])
    bucket1 = Bucket(auth=auth1, endpoint=pickle_file_key_sts['endpoint'], bucket_name=pickle_file_key_sts['bucket'],
                     region=pickle_file_key_sts['region'])
    with TqdmUpTo(desc=f"Registering Func...") as t:
        bucket1.put_object(pickle_file_key_sts['fileKey'], func_source, progress_callback=t.update_to)
        t.total = t.n

    auth2 = StsAuth(access_key_id=meta_file_key_sts['accessKeyId'],
                    access_key_secret=meta_file_key_sts['accessKeySecret'],
                    security_token=meta_file_key_sts['securityToken'])
    bucket2 = Bucket(auth=auth2, endpoint=meta_file_key_sts['endpoint'], bucket_name=meta_file_key_sts['bucket'],
                     region=meta_file_key_sts['region'])
    with TqdmUpTo(desc=f"Registering Meta...") as t:
        bucket2.put_object(meta_file_key_sts['fileKey'], get_meta(func_source), progress_callback=t.update_to)
        t.total = t.n

    return func_uuid



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import abc
import inspect
from typing import Union
from types import FunctionType

from aiearth import engine
from aiearth.engine.variable_node import VariableNode
from aiearth.engine.function_node import FunctionNode
from aiearth.engine.customfunction_node import CustomFunctionNode
from aiearth.engine.function_helper import FunctionHelper
from aiearth.core.error.aie_error import AIEError, AIEErrorCode


class List(FunctionNode):
    def __init__(self, lst: object) -> engine.List:
        if lst is not None and not isinstance(lst, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"lst 只支持object类型参数, 传入类型为{type(lst)}"
            )

        invoke_args = {
            "lst": lst,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "lst" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数lst不能为空")

        super(List, self).__init__("List.constructors", invoke_args)

    def add(self, element: object) -> engine.List:
        if element is not None and not isinstance(element, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"element 只支持object类型参数, 传入类型为{type(element)}"
            )

        invoke_args = {
            "input": self,
            "element": element,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "element" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数element不能为空")

        return FunctionHelper.apply("List.add", "engine.List", invoke_args)

    def get(self, index: int) -> engine.List:
        if index is not None and not isinstance(index, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"index 只支持int类型参数, 传入类型为{type(index)}"
            )

        invoke_args = {
            "input": self,
            "index": index,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "index" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数index不能为空")

        return FunctionHelper.apply("List.get", "engine.List", invoke_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .function_node import FunctionNode


class VariableNode(FunctionNode):
    def __init__(self, var_name):
        super(VariableNode, self).__init__(None, None, var_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from aiearth.core.error import AIEError, AIEErrorCode
from aiearth.core.client.client import BaseClient
from aiearth.core.client.endpoints import Endpoints, MarianaDataResource
import oss2
import urllib


class Dataset(BaseClient):
    def __init__(self, name):
        self.name = name
        self.file_entries = []
        self.dataset_id = -1

    class StatusCode():
        PUBLISHING = 0
        COMPLETED = 6
        FAILED = -1

    class Status():
        PUBLISHING = "发布中"
        COMPLETED = "发布完成"
        FAILED = "发布失败"

    def upload(self, file, suffix):
        headers = {"Content-Type": "application/json"}

        get_sts_token_url = Endpoints.MARIANA_DATA + \
            MarianaDataResource.GET_STS_TOKEN.format(suffix)
        response = super(Dataset, self).get(
            get_sts_token_url, headers).json()
        if response["code"] != 0:
            raise AIEError(response["code"], response["message"])

        ak = response["module"]["accessKeyId"]
        sk = response["module"]["accessKeySecret"]
        token = response["module"]["securityToken"]
        region = response["module"]["region"]
        bucket = response["module"]["bucket"]
        file_key = response["module"]["fileKey"]
        file_name = os.path.basename(file)

        auth = oss2.StsAuth(ak, sk, token)
        oss_bucket = oss2.Bucket(
            auth, "http://{}-internal.aliyuncs.com".format(region), bucket)
        headers = {
            "Content-Disposition": f"attachment;filename={urllib.parse.quote(file_name)}"}
        oss_bucket.put_object_from_file(file_key, file, headers)

        self.file_entries.append({
            "name": file_name,
            "uri": "oss://{}/{}".format(bucket, file_key)
        })

    def publish(self):
        headers = {"Content-Type": "application/json"}

        publish_url = Endpoints.MARIANA_DATA + MarianaDataResource.PUBLISH
        payload = {
            "datasets": [{"name": self.name, "entries": self.file_entries}]
        }
        response = super(Dataset, self).post(
            publish_url, headers, payload).json()
        if response["code"] != 0:
            raise AIEError(response["code"], response["message"])

        self.dataset_id = response["module"][0]

    def __toStatus(self, status):
        if status == Dataset.StatusCode.COMPLETED:
            return Dataset.Status.COMPLETED
        elif status == Dataset.StatusCode.PUBLISHING:
            return Dataset.Status.PUBLISHING
        else:
            return Dataset.Status.FAILED

    def status(self):
        if self.dataset_id == -1:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "dataset_id未初始化")

        headers = {"Content-Type": "application/json"}

        status_url = Endpoints.MARIANA_DATA + \
            MarianaDataResource.GET_PUBLISH_STATUS.format(self.dataset_id)
        response = super(Dataset, self).get(
            status_url, headers).json()
        if response["code"] != 0:
            raise AIEError(response["code"], response["message"])

        status = response["module"]["publishStatus"]
        return self.__toStatus(status)

    def active(self):
        status = self.status()
        return status == Dataset.Status.PUBLISHING



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import inspect

from aiearth.engine.aie_object.list import List

from aiearth import engine
import collections
from aiearth.core.error import AIEError, AIEErrorCode
from types import FunctionType
from aiearth.engine.serialize.udf_manager import manage
from aiearth.engine.udf_base.aie_udf_base import AIEUDF
from aiearth.engine.udf_base.aie_udaf_base import AIEUDAF


class GenVisitor(object):
    def __init__(self, is_compacted=True):
        self.is_compacted = is_compacted
        # {obj_id: ref}
        self.refs = {}
        # [(ref, value)]
        self.scope = []
        # {ref: use_count}
        self.ref_uses_cnt = collections.defaultdict(int)

    def visit(self, obj):
        obj_id = id(obj)
        if self.is_compacted:
            ref = self.refs.get(obj_id, None)
            if ref:
                self.ref_uses_cnt[ref] += 1
                return ref

        if hasattr(obj, "iterables"):
            result = {
                    "argumentReference": getattr(obj, "iterables")
                }
        elif isinstance(obj, engine.customfunction_node.CustomFunctionNode):
            result = {
                "functionDefinitionValue": {
                    "argumentNames": obj.arg_names,
                    "body": self.visit(obj.body)
                }
            }
        elif isinstance(obj, engine.function_node.FunctionNode):
            if obj._isVariable():
                result = {
                    "argumentReference": obj.var_name
                }
            else:
                args_dict = dict()
                for key in sorted(obj.invoke_args):
                    args_dict[key] = self.visit(obj.invoke_args[key])
                if self.is_compacted:
                    result = {
                        "functionInvocationValue": {
                            "functionName": obj.func_name,
                            "arguments": {k: {
                                "valueReference": v
                            } for k, v in args_dict.items()},
                        }
                    }
                else:
                    result = {
                        "functionInvocationValue": {
                            "functionName": obj.func_name,
                            "arguments": args_dict,
                        }
                    }
        elif isinstance(obj, (int, float, bool, str)):
            result = {"constantValue": obj}
        elif isinstance(obj, (list, tuple)):
            values = [self.visit(item) for item in obj]
            """ constant optimizer
            "arrayValue": {
                "values": [
                    {
                        "constantValue": "xx"
                    }
                ]
            }
            """
            if all("constantValue" in value for value in values):
                result = {
                    "constantValue": [value["constantValue"] for value in values]
                }
            else:
                if self.is_compacted:
                    result = {"arrayValue": {
                        "values": [{"valueReference": value} for value in values]
                    }}
                else:
                    result = {"arrayValue": {
                        "values": values
                    }}
        elif isinstance(obj, dict):
            values = {k: self.visit(obj[k]) for k in sorted(obj)}
            """ constant optimizer
            "dictionaryValue": {
                "values": {
                    "key1": {
                        "constantValue": "xx"
                    }
                }
            }
            """
            if all("constantValue" in value for value in values.values()):
                result = {
                    "constantValue": {key: value["constantValue"] for key, value in values.items()}
                }
            else:
                if self.is_compacted:
                    result = {
                        "dictionaryValue": {
                            "values": {
                                k: {
                                    "valueReference": v
                                } for k, v in values.items()
                            }
                        }
                    }
                else:
                    result = {
                        "dictionaryValue": {
                            "values": values
                        }
                    }
        elif hasattr(obj, "udf") and getattr(obj, 'udf') == 'true' and issubclass(getattr(obj, 'clz'), AIEUDF):
            func_uuid = manage(getattr(obj, 'clz'))
            result = {"constantValue": func_uuid}
        elif hasattr(obj, "udaf") and getattr(obj, "udaf") == "true" and issubclass(getattr(obj, 'clz'), AIEUDAF):
            func_uuid = manage(getattr(obj, "clz"))
            result = {"constantValue": func_uuid}
        else:
            raise AIEError(AIEErrorCode.DEFAULT_INTERNAL_ERROR,
                                               "", "serialize.GenVisitor.visit not support obj type %s" % type(
                                                   obj)
                                               )

        if self.is_compacted:
            ref = self.refs.get(obj_id, None)
            if not ref:
                ref = str(len(self.scope))
                self.scope.append((ref, result))
                self.refs[obj_id] = ref
            self.ref_uses_cnt[ref] += 1
            return ref
        else:
            return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import abc
import inspect
from typing import Union
from types import FunctionType

from aiearth import engine
from aiearth.engine.variable_node import VariableNode
from aiearth.engine.function_node import FunctionNode
from aiearth.engine.customfunction_node import CustomFunctionNode
from aiearth.engine.function_helper import FunctionHelper
from aiearth.core.error.aie_error import AIEError, AIEErrorCode


class Collection(FunctionNode):
    @abc.abstractmethod
    def elementType(self):
        pass

    def filter(self, filter: Union[str, engine.Filter]) -> engine.Collection:
        if filter is not None and not isinstance(filter, (str, engine.Filter)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"filter 只支持(str,engine.Filter)类型参数, 传入类型为{type(filter)}",
            )

        invoke_args = {
            "collection": self,
            "filter": filter,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "filter" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数filter不能为空")

        return FunctionHelper.apply(
            "Collection.filter", "engine.Collection", invoke_args
        )

    def filterBounds(
        self, geometry: Union[engine.Geometry, engine.Feature, engine.FeatureCollection]
    ) -> engine.Collection:
        if geometry is not None and not isinstance(
            geometry, (engine.Geometry, engine.Feature, engine.FeatureCollection)
        ):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"geometry 只支持(engine.Geometry,engine.Feature,engine.FeatureCollection)类型参数, 传入类型为{type(geometry)}",
            )

        invoke_args = {
            "collection": self,
            "geometry": geometry,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "geometry" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数geometry不能为空")

        return FunctionHelper.apply(
            "Collection.filterBounds", "engine.Collection", invoke_args
        )

    def first(self) -> object:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Collection.first", "object", invoke_args)

    def limit(
        self, limit: int, property: str = None, ascending: bool = True
    ) -> engine.Collection:
        if limit is not None and not isinstance(limit, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"limit 只支持int类型参数, 传入类型为{type(limit)}"
            )

        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        if ascending is not None and not isinstance(ascending, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"ascending 只支持bool类型参数, 传入类型为{type(ascending)}",
            )

        invoke_args = {
            "collection": self,
            "limit": limit,
            "property": property,
            "ascending": ascending,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "limit" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数limit不能为空")

        return FunctionHelper.apply(
            "Collection.limit", "engine.Collection", invoke_args
        )

    def size(self) -> object:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Collection.size", "object", invoke_args)

    def map(self, baseAlgorithm) -> engine.Collection:
        import types

        if baseAlgorithm is not None and not isinstance(
            baseAlgorithm, types.FunctionType
        ):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"baseAlgorithm 只支持function类型参数, 传入类型为{type(baseAlgorithm)}",
            )

        args = inspect.getfullargspec(baseAlgorithm).args
        mapping_args = [
            "_MAPPING_VAR_#arg" + str(k) + "_" + v for k, v in enumerate(args)
        ]
        variables = [VariableNode(mapping_arg) for mapping_arg in mapping_args]

        element_type = self.elementType()

        def func_warp(e):
            return baseAlgorithm(element_type(e))

        body = func_warp(*variables)
        if body is None:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "自定义map函数必须有返回值")
        customfunc = CustomFunctionNode(mapping_args, body)

        invoke_args = {
            "collection": self,
            "baseAlgorithm": customfunc,
        }
        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "baseAlgorithm" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数baseAlgorithm不能为空")

        return FunctionHelper.apply("Collection.map", "engine.Collection", invoke_args)

    def iterate(self, func: FunctionType, first: object) -> object:
        if func is not None and not isinstance(func, FunctionType):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"func 只支持FunctionType类型参数, 传入类型为{type(func)}"
            )

        if first is not None and not isinstance(first, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"first 只支持object类型参数, 传入类型为{type(first)}"
            )

        args = inspect.getfullargspec(func).args
        mapping_args = [
            "_MAPPING_VAR_#arg" + str(k) + "_" + v for k, v in enumerate(args)
        ]
        variables = [VariableNode(mapping_arg) for mapping_arg in mapping_args]

        element_type = self.elementType()

        def func_warp(collection_element_self, param_object: object):
            return func(element_type(collection_element_self), param_object)

        from copy import deepcopy

        deepcopy_first = deepcopy(first)
        setattr(deepcopy_first, "iterables", variables[1].var_name)
        body = func_warp(variables[0], deepcopy_first)
        if body is None:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "自定义map函数必须有返回值")
        customfunc = CustomFunctionNode(mapping_args, body)

        invoke_args = {
            "collection": self,
            "baseAlgorithm": customfunc,
            "first": first,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "baseAlgorithm" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数func不能为空")

        if "first" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数first不能为空")

        return FunctionHelper.apply("Collection.iterate", "object", invoke_args)

    def remap(
        self, lookupIn: list, lookupOut: list, columnName: str
    ) -> engine.Collection:
        if lookupIn is not None and not isinstance(lookupIn, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"lookupIn 只支持list类型参数, 传入类型为{type(lookupIn)}"
            )

        if lookupOut is not None and not isinstance(lookupOut, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"lookupOut 只支持list类型参数, 传入类型为{type(lookupOut)}",
            )

        if columnName is not None and not isinstance(columnName, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"columnName 只支持str类型参数, 传入类型为{type(columnName)}",
            )

        invoke_args = {
            "input": self,
            "lookupIn": lookupIn,
            "lookupOut": lookupOut,
            "columnName": columnName,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "lookupIn" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数lookupIn不能为空")

        if "lookupOut" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数lookupOut不能为空")

        if "columnName" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数columnName不能为空")

        return FunctionHelper.apply(
            "Collection.remap", "engine.Collection", invoke_args
        )
